from django.db import (
    NotSupportedError
)


class InefficientQueryError(NotSupportedError):
    message = (
        'Innefficent queries are not allowed on this model. '
        'Make sure if you are filtering that you are only '
        'filtering on either primary key or indexed fields '
        'and that if you are ordering that you are only '
        'ordering on fields that are components of compound '
        'primary keys.\n\nAlternatively you can enable '
        'inefficient filtering and ordering in your database '
        'by setting the ALLOW_INEFFICIENT_QUERIES=True in '
        'your settings.py or you can set this per model in '
        'the Cassandra meta class for you model '
        '"allow_inefficient_queries=True".'
    )

    def __init__(self, query):
        self.query = query

    def __str__(self):
        return  (self.message + ':\n%s') % (repr(self.query),)


class InvalidQueryOpException(NotSupportedError):
    message = (
        'The operation %s is not supported.'
    )

    def __init__(self, operation):
        self.operation = operation

    def __str__(self):
        return (self.message + ':\n%s') % (repr(self.operation),)



from cassandra.cqlengine import management as db_management
from cassandra.cqlengine.columns import (
    Map,
    Text
)

from django.db.backends.schema import (
    BaseDatabaseSchemaEditor
)

from djangocassandra.db.meta import (
    CqlColumnFamilyMetaClass,
    get_column_family,
    internal_type_to_column_map
)


class CassandraSchemaEditor(BaseDatabaseSchemaEditor):
    known_models = set()

    def skip_default(
        self,
        field
    ):
        return True

    def _create_db_table(
        self,
        column_family
    ):
        db_management.sync_table(column_family)

    def create_model(
        self,
        model
    ):
        self.connection.create_keyspace()

        column_family = get_column_family(
            self.connection,
            model
        )
        self._create_db_table(column_family)

    def delete_model(
        self,
        model
    ):
        column_family = get_column_family(
            self.connection,
            model
        )

        db_management.drop_table(column_family)

    def alter_unique_together(
        self,
        model,
        old_unique_together,
        new_unique_together
    ):
        '''
        Unique together isn't a concept Cassandra understands
        out of the box. We will just do a noop in this case.
        '''
        pass

    def alter_index_together(
        self,
        model,
        old_index_together,
        new_index_together
    ):
        '''
        Index together is a similar concept to Cassandra's compound
        primary keys but to maintain compatibilty with Django projects
        that are migrating from some other database backend I will keep
        these concepts seperate instead of trying to overload
        _meta.index_together
        '''
        pass

    def alter_db_table(
        self,
        model,
        old_db_table,
        new_db_table
    ):
        '''
        Renaming the table isn't supported by Cassandra. This
        has to be done in a manual process as follows:

            1) create schema for NEW_Keyspace
            2) stop writes to OLD_Keyspace from app (reads can continue)
            3) flush OLD_Keyspace on every node, via nodetool
            4) hard link all sstables from OLD_Keyspace directory to
               NEW_Keyspace directory
            5) call nodetool -h localhost refresh NEW_Keyspace
            6) enable reads/writes from/to NEW_Keyspace from app
               (disable reads on OLD_Keyspace)
            7) clean up OLD_Keyspace (drop schema, delete files, etc.)
        '''
        if (
            new_db_table._meta.db_table !=
            old_db_table._meta.db_table
        ):
            raise Exception('Renaming models not supported')

        self.create_model(new_db_table)

    def alter_db_tablespace(
        self,
        model,
        old_db_tablespace,
        new_db_tablespace
    ):
        '''
        Moving a model to a different keyspace requires
        creating the column family in the new keyspace
        (new_db_tablespace), migrating the data and then
        deleting the old column family.
        '''
        pass

    def add_field(
        self,
        model,
        field
    ):
        self.create_model(model)

    def remove_field(
        self,
        model,
        field
    ):
        self.create_model(model)

    def alter_field(
        self,
        model,
        old_field,
        new_field,
        strict=False
    ):
        self.create_model(model)



from djangotoolbox.db.creation import NonrelDatabaseCreation

from cassandra.cqltypes import (
    UUIDType,
    LongType,
    BooleanType,
    VarcharType,
    DateType,
    TimestampType,
    DecimalType,
    FloatType,
    IntegerType,
    InetAddressType,
    Int32Type,
    UTF8Type,
    ListType,
    SetType,
    MapType,
    BytesType
)


class DatabaseCreation(NonrelDatabaseCreation):
    data_typename_to_typeclass = {
        UUIDType.typename: UUIDType,
        LongType.typename: LongType,
        BooleanType.typename: BooleanType,
        VarcharType.typename: VarcharType,
        DateType.typename: DateType,
        TimestampType.typename: TimestampType,
        DecimalType.typename: DecimalType,
        FloatType.typename: FloatType,
        IntegerType.typename: IntegerType,
        InetAddressType.typename: InetAddressType,
        Int32Type.typename: Int32Type,
        UTF8Type.typename: UTF8Type,
        ListType.typename: ListType,
        SetType.typename: SetType,
        MapType.typename: MapType,
        BytesType.typename: BytesType
    }

    data_types = {
        # NoSQL databases often have specific concepts of entity keys.
        # For example, GAE has the db.Key class, MongoDB likes to use
        # ObjectIds, Redis uses strings, while Cassandra supports
        # different types (including binary data).
        'AutoField': UUIDType.typename,
        'RelatedAutoField': UUIDType.typename,
        'ForeignKey': UUIDType.typename,
        'OneToOneField': UUIDType.typename,
        'ManyToManyField': UUIDType.typename,

        # Standard field types, more or less suitable for a database
        # (or its client / driver) being able to directly store or
        # process Python objects.
        'BigIntegerField': LongType.typename,
        'BooleanField': BooleanType.typename,
        'CharField': VarcharType.typename,
        'CommaSeparatedIntegerField': VarcharType.typename,
        'DateField': DateType.typename,
        'DateTimeField': TimestampType.typename,
        'DecimalField': DecimalType.typename,
        'EmailField': VarcharType.typename,
        'FileField': VarcharType.typename,
        'FilePathField': VarcharType.typename,
        'FloatField': FloatType.typename,
        'ImageField': VarcharType.typename,
        'IntegerField': IntegerType.typename,
        'IPAddressField': InetAddressType.typename,
        'NullBooleanField': BooleanType.typename,
        'PositiveIntegerField': IntegerType.typename,
        'PositiveSmallIntegerField': Int32Type.typename,
        'SlugField': VarcharType.typename,
        'SmallIntegerField': Int32Type.typename,
        'TextField': UTF8Type.typename,
        'TimeField': TimestampType.typename,
        'URLField': VarcharType.typename,
        'BinaryField': BytesType.typename,
        # You may use "list" for SetField, or even DictField and
        # EmbeddedModelField (if your database supports nested lists).
        # All following fields also support "string" and "bytes" as
        # their storage types -- which work by serializing using pickle
        # protocol 0 or 2 respectively.
        # Please note that if you can't support the "natural" storage
        # type then the order of field values will be undetermined, and
        # lookups or filters may not work as specified (e.g. the same
        # set or dict may be represented by different lists, with
        # elements in different order, so the same two instances may
        # compare one way or the other).
        'AbstractIterableField': ListType.typename,
        'ListField': ListType.typename,
        'SetField': SetType.typename,
        'DictField': MapType.typename,
        'EmbeddedModelField': MapType.typename,
        # RawFields ("raw" db_type) are used when type is not known
        # (untyped collections) or for values that do not come from
        # a field at all (model info serialization), only do generic
        # processing for them (if any). On the other hand, anything
        # using the "bytes" db_type should be converted to a database
        # blob type or stored as binary data.
        'RawField': BytesType.typename,
        'BlobField': BytesType.typename
    }

    def db_type(
        self,
        field
    ):
        '''
        TODO:
        Investigate wether this code is redundant.
        After looking at Field.db_type() it seems like
        the code here is duplicated.
        '''
        data_type = self.data_types.get(field.get_internal_type())

        if None is data_type:
            return field.db_type(connection=self.connection)

        else:
            return data_type

    def sql_create_model(
        self,
        model,
        style,  # Used for styling output
        known_models=set()
    ):
        meta = model._meta

        if (
            not meta.managed or
            meta.proxy or
            meta.db_table in known_models
        ):
            return [], {}

        schema_editor = self.connection.schema_editor()
        schema_editor.create_model(model)

        return [], {}

    def _create_test_db(
        self,
        verbosity,
        autoclobber,
        keepdb=False
    ):
        test_database_name = self._get_test_db_name()

        qn = self.connection.ops.quote_name

        nodb_connection = self._nodb_connection

        nodb_connection.create_keyspace(
            qn(test_database_name)
        )

        return test_database_name



from django.db.models.query import QuerySet as DjangoQuerySet


class QuerySet(DjangoQuerySet):
    def next(self, limit=None):
        last_limit = len(self)

        if 0 == last_limit:
            return self.model.objects.none()

        last = self[last_limit - 1]

        new_set = self.model.objects.all()
        new_set.where = self.query.where
        new_set = new_set.filter(pk_token__gt=last.pk_token)

        if None is not limit:
            new_set = new_set[:limit]

        else:
            new_set = new_set[:last_limit]

        return new_set



import uuid

from math import log10, floor
from datetime import datetime
from django.utils.six import with_metaclass
from django.db.models import (
    Field,
    AutoField,
    SubfieldBase,
    CharField,
    ForeignKey,
    DateTimeField as DjangoDateTimeField
)
from django.utils.translation import ugettext_lazy as _

from cassandra.cqlengine.functions import Token

from .values import PrimaryKeyValue


class DateTimeField(DjangoDateTimeField):
    def get_prep_value(self, value):
        # Hack cassandra truncates microseconds to milliseconds.
        microsecond = value.microsecond
        if 0 != microsecond:
            microsecond = round(
                value.microsecond,
                -int(floor(log10(abs(value.microsecond)))) + 2
            )
            value = datetime(
                value.year,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
                int(microsecond),
                value.tzinfo
            )

        return super(DateTimeField, self).get_prep_value(value)


class TokenPartitionKeyField(Field):
    def __init__(
        self,
        *args,
        **kwargs
    ):
        update_kwargs = {
            'null': True,
            'blank': True,
            'editable': False,
            'db_column': 'pk__token'
        }

        kwargs.update(update_kwargs)

        super(TokenPartitionKeyField, self).__init__(
            *args,
            **kwargs
        )

    def contribute_to_class(
        self,
        cls,
        name
    ):
        super(TokenPartitionKeyField, self).contribute_to_class(
            cls,
            name
        )

        setattr(cls, name, self)

    def get_internal_type(self):
        return 'Token'

    def __get__(
        self,
        instance,
        instance_type=None
    ):
        if None is instance:
            return self

        if (
            hasattr(instance, 'Cassandra') and
            hasattr(instance.Cassandra, 'partition_keys') and
            1 < len(instance.Cassandra.partition_keys)
        ):
            compound_pk = []
            for key in instance.Cassandra.partition_keys:
                field = instance._meta.get_field(key)
                if isinstance(field, ForeignKey):
                    key += '_id'

                value = getattr(instance, key)
                try:
                    value = uuid.UUID(value)

                except:
                    pass

                compound_pk.append(value)

            pk = tuple(compound_pk)

        else:
            pk = instance.pk
            try:
                pk = uuid.UUID(pk)

            except:
                pass

        return Token(pk)

    def __set__(
        self,
        instance,
        value
    ):
        pass

    def value_to_string(
        self,
        obj
    ):
        value = self._get_val_from_obj(obj)
        return ''.join([
            'token(',
            ','.join(str(value.value)),
            ')'
        ])


class FieldUUID(with_metaclass(SubfieldBase, CharField)):
    description = _('UUID')

    default_error_messages = {
        'invalid': _("'%(value)s' value must be a valid UUID."),
    }

    def __init__(
        self,
        *args,
        **kwargs
    ):
        if 'default' not in kwargs:
            kwargs['default'] = uuid.uuid4

        kwargs['max_length'] = 36

        super(FieldUUID, self).__init__(
            *args,
            **kwargs
        )

    def to_python(
        self,
        value
    ):
        if isinstance(value, uuid.UUID):
            return str(value)

        else:
            return value

    def get_prep_value(self, value):
        value = super(FieldUUID, self).get_prep_value(value)
        if (
            value is None or
            isinstance(value, uuid.UUID)
        ):
            return value

        try:
            return uuid.UUID(value)

        except:
            return value

    def get_internal_type(self):
        return 'FieldUUID'

    @staticmethod
    def get_auto_value(self):
        return uuid.uuid4()

    def value_to_string(self, value):
        if isinstance(value, basestring):
            return value

        try:
            return str(value)

        except (TypeError, ValueError):
            raise Exception(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )


class AutoFieldUUID(with_metaclass(SubfieldBase, AutoField)):
    description = _('UUID')

    default_error_messages = {
        'invalid': _("'%(value)s' value must be a valid UUID."),
    }

    def __init__(
        self,
        *args,
        **kwargs
    ):
        if 'default' not in kwargs:
            kwargs['default'] = uuid.uuid4

        super(AutoFieldUUID, self).__init__(
            *args,
            **kwargs
        )

    def to_python(
        self,
        value
    ):
        if isinstance(value, uuid.UUID):
            return str(value)

        else:
            return value

    def get_prep_value(self, value):
        if (
            value is None or
            isinstance(value, uuid.UUID)
        ):
            return value

        try:
            return uuid.UUID(value)

        except:
            return value

    def get_internal_type(self):
        return 'AutoFieldUUID'

    @staticmethod
    def get_auto_value(self):
        return uuid.uuid4()

    def value_to_string(self, value):
        if isinstance(value, basestring):
            return value

        try:
            return str(value)

        except (TypeError, ValueError):
            raise Exception(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )


class PrimaryKeyField(Field):
    def __init__(
        self,
        *args,
        **kwargs
    ):
        self.field_class = kwargs.pop(
            'field_class',
            AutoFieldUUID
        )
        self.field_args = args
        self.field_kwargs = kwargs

        super(PrimaryKeyField, self).__init__()

    def contribute_to_class(
        self,
        cls,
        name
    ):
        self.name = name

        if (
            hasattr(cls, "Cassandra") and
            hasattr(cls.Cassandra, "partition_keys")
        ):
            if hasattr(cls.Cassandra, "clustering_keys"):
                self.all_keys = [key for key in chain(
                    cls.Cassandra.partition_keys,
                    cls.Cassandra.clustering_keys
                )]

            else:
                self.all_keys = cls.Cassandra.partition_keys

        class PrimaryKeyField(self.field_class):
            def __init__(self, *args, **kwargs):
                kwargs["primary_key"] = True

                super(PrimaryKeyField, self).__init__(
                    *args,
                    **kwargs
                )

            def get_prep_value(self, value):
                if isinstance(value, PrimaryKeyValue):
                    return value

                elif isinstance(self, ForeignKey):
                    return self.related_field.get_prep_value(value)
                    
                else:
                    return super(PrimaryKeyField, self).get_prep_value(value)

            def get_internal_type(self):
                if isinstance(self, ForeignKey):
                    return "ForeignKey"

                return super(
                    PrimaryKeyField,
                    self
                ).get_internal_type()

        cls.add_to_class(
            name,
            PrimaryKeyField(
                *self.field_args,
                **self.field_kwargs
            )
        )



try:
    import cassandra as Database

except ImportError as e:
    from django.core.exceptions import ImproperlyConfigured
    raise ImproperlyConfigured(
        "Error loading cassandra-driver module (python-driver): %s" % e
    )


class CassandraCursor(object):
    """
    A wrapper around Datastax Cassandra Python Driver Session class to
    make it behave like a database cursor in a relational database such as
    MySQL and so that we can catch particular exception instances and reraise
    them with the right types. Implemented as a wrapper, rather than a
    subclass, so that we aren't stuck to the particular underlying
    representation returned by Connection.cursor().
    """
    def __init__(self, session):
        self._reset_attributes()
        self.session = session

    def _reset_attributes(self):
        self.session = None
        self.rows = []
        self.index = -1
        self._last_statement = None
        self._with_rows = False
        self._lastrowid = None

    @property
    def lastrowid(self):
        return self._lastrowid

    @property
    def with_rows(self):
        return self._with_rows

    @property
    def statement(self):
        return self._last_statement

    @property
    def column_name(self):
        raise NotImplementedError(
            'column_name propery has not been implemented'
        )

    @property
    def description(self):
        raise NotImplementedError(
            'description propery has not been implemented'
        )

    def fetchone(self):
        '''
        Note that this will have interesting behavior since there is
        no efficent way to check if we are at the last result, perhaps
        if the current result = first result?
        '''
        self.index += 1
        return self.rows[self.index - 1]

    def fetchmany(
        self,
        size=1
    ):
        many = self.rows[self.index:self.index + size]
        self.index += size

        return many

    def fetchall(self):
        remaining = self.rows[self.index:]
        self.rows = []
        self.index = -1

        return remaining

    def execute(self, query, args=None):
        self.rows = self.session.execute(query, args)
        self.index = 0
        if hasattr(self.rows, 'current_rows'):
            self._with_rows = (
                None is not self.rows and len(
                    self.rows.current_rows
                )
            )

        else:
            self._with_rows = (
                None is not self.rows and len(self.rows)
            )

        return self.rows

    def executemany(self, query, args_list):
        prepared = self.session.prepare(query)
        batch = Database.BatchStatement()
        for args in args_list:
            batch.add(prepared, args)

        self.rows = self.session.execute(batch)
        self.index = 0
        self._with_rows = 0 != len(self.rows)

        return self.rows

    def rollback(self):
        pass

    def commit(self):
        pass

    def close(self):
        self._reset_attributes()

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            return getattr(self.session, attr)

    def __iter__(self):
        return self.rows

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



from collections import OrderedDict


class PrimaryKeyValue(OrderedDict):
    def __int__(self):
        return self.__hash__()

    def __hash__(self):
        return hash(self.to_tuple())

    def __str__(self):
        return str(self.to_tuple())

    def __unicode__(self):
        return unicode(self.to_tuple())

    def to_tuple(self):
        return tuple(self.iteritems())




import itertools

from django.db.backends import BaseDatabaseIntrospection
from djangotoolbox.db.base import NonrelDatabaseIntrospection


class DatabaseIntrospection(NonrelDatabaseIntrospection):
    def get_table_list(
        self,
        cursor=None
    ):
        if None is cursor:
            cursor = self.connection.cursor()

        current_keyspace = cursor.keyspace

        keyspaces = [
            key for key in self.connection.settings_dict.get(
                'KEYSPACES'
            ).keys()
        ]

        table_list = []
        '''
        These are where the schema information for the cluster is
        stored for Cassandra version 2.x
        '''
        schema_keyspace = 'system'
        schema_table_name = 'schema_columnfamilies'
        schema_table_name_field = 'columnfamily_name'

        if 'system_schema' in self.connection.cluster.metadata.keyspaces:
            '''
            If there is a 'system_schema' keyspace then we are on
            Cassandra 3.x and need to look for the schema info here.
            '''
            schema_keyspace = 'system_schema'
            schema_table_name = 'tables'
            schema_table_name_field = 'table_name'

        try:
            cursor.set_keyspace(schema_keyspace)
            for keyspace in keyspaces:
                table_list = itertools.chain(
                    table_list, cursor.execute(
                        ''.join([
                            'SELECT ',
                            schema_table_name_field,
                            ' from ',
                            schema_table_name,
                            ' where keyspace_name=\'',
                            keyspace,  # TODO: SANITIZE ME JUST IN CASE!!!
                            '\''
                        ])
                    )
                )

        finally:
            if None is not current_keyspace:
                cursor.set_keyspace(current_keyspace)

        return [row[schema_table_name_field] for row in table_list]

    def table_names(self, cursor=None):
        return BaseDatabaseIntrospection.table_names(self, cursor)



from distutils.core import setup

setup(
    name='djangocassandra',
    version='0.7.7',
    description='Cassandra support for the Django web framework',
    long_description=(
        'The Cassandra database backend for Django has been '
        'developed to allow developers to utilize the Apache '
        'Cassandra Database in their Django projects. '
        'Read the full documentation at '
        'https://djangocassandra.readthedocs.org'
    ),
    author='Seth Denner',
    author_email='seth@knotis.com',
    maintainer='Knotis Inc.',
    maintainer_email='support@knotis.com',
    url='https://github.com/Knotis/djangocassandra',
    download_url=(
        'https://github.com/Knotis/djangocassandra/archive/master.zip'
    ),
    license='BSD License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django :: 1.7',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
        'Topic :: Internet',

    ],
    packages=[
        'djangocassandra',
        'djangocassandra.db',
        'djangocassandra.db.backends',
        'djangocassandra.db.backends.cassandra'
    ],
    install_requires=[
        'django>=1.7, < 1.8',
        'cassandra-driver==3.7.0',
        'blist',
        'djangotoolbox==1.7.0'
    ],
)



from itertools import chain
from collections import OrderedDict

from django.apps import apps
from django.db.models import (
    Model as DjangoModel,
    Manager,
    ForeignKey
)
from django.db.models.fields import (
    FieldDoesNotExist
)

from .fields import TokenPartitionKeyField
from .values import PrimaryKeyValue
from .query import QuerySet


class ColumnFamilyManager(Manager.from_queryset(QuerySet)):
    denormalized_models = []

    @staticmethod
    def denormalize(
        origin,
        destination
    ):
        field_names = origin._meta.get_all_field_names()
        for name in field_names:
            try:
                field = origin._meta.get_field(name)

            except FieldDoesNotExist:
                continue

            setattr(
                destination,
                field.name,
                getattr(
                    origin,
                    field.name
                )
            )

        return destination

    def create(
        self,
        *args,
        **kwargs
    ):
        instance = super(ColumnFamilyManager, self).create(
            *args,
            **kwargs
        )

        return instance



class ColumnFamilyModel(DjangoModel):
    class Meta:
        abstract = True

    pk_token = TokenPartitionKeyField()

    def __init__(
        self,
        *args,
        **kwargs
    ):
        super(ColumnFamilyModel, self).__init__(
            *args,
            **kwargs
        )

    def __hash__(self):
        pk_value = self._get_pk_val()

        if isinstance(pk_value, dict):
            return int(pk_value)

        else:
            return super(
                ColumnFamilyModel,
                self
            ).__hash__()
        

    def _get_pk_val(self, meta=None):
        if (
            hasattr(self, "Cassandra") and
            hasattr(self.Cassandra, "partition_keys")
        ):
            if hasattr(self.Cassandra, "clustering_keys"):
                all_keys = [key for key in chain(
                    self.Cassandra.partition_keys,
                    self.Cassandra.clustering_keys
                )]

            else:
                all_keys = self.Cassandra.partition_keys

            if 1 < len(all_keys):
                pk_value = PrimaryKeyValue()
                for key in all_keys:
                    field = self._meta.get_field_by_name(key)[0]
                    if isinstance(field, ForeignKey):
                        key += "_id"

                    pk_value[key] = field.get_prep_value(getattr(
                        self,
                        key
                    ))

                return pk_value

        return super(
            ColumnFamilyModel,
            self
        )._get_pk_val(meta=meta)
                
    def _set_pk_val(self, value):
        if (
            hasattr(self, "Cassandra") and
            hasattr(self.Cassandra, "partition_keys")
        ):
            if hasattr(self.Cassandra, "clustering_keys"):
                all_keys = [field for field in chain(
                    self.Cassandra.partition_keys,
                    self.Cassandra.clustering_keys
                )]

            else:
                all_keys = self.Cassandra.partition_keys

            if 1 < len(all_keys):
                if not isinstance(value, dict):
                    raise self.PrimaryKeyInconsistencyException(
                        "PK values must be a dict of the form "
                        "{ field: value, ... }"
                    )

                items = [i for i in value.iteritems()]

                if len(items) != len(pk_fields):
                    raise self.PrimaryKeyInconsistencyException(
                        "Expected %s field/value pairs, recieved %s." % (
                            len(pk_fields),
                            len(items)
                        )
                    )

                for field, value in items:
                    if field not in pk_fields:
                        raise self.PrimaryKeyInconsistencyException((
                            "Field \"%s\" is not part of the primary key."
                            "must be one of %s"
                        ) % (field, pk_fields))

                # Loop twice to avoid partialy updating the key.
                for field, value in items:
                    setattr(self, field, value)

                return None

        return super(
            ColumnFamilyModel,
            self
        )._set_pk_val(value)

    pk = property(_get_pk_val, _set_pk_val)

    @staticmethod
    def should_denormalize(instance):
        '''
        Return false if you don't want to denormalize the
        passed in instance into this column family.
        '''
        return True

    def denormalize(
        self,
        instance
    ):
        ColumnFamilyManager.denormalize(
            instance,
            self
        )

    def save(
        self,
        *args,
        **kwargs
    ):
        kwargs['force_insert'] = True
        kwargs['force_update'] = False

        super(ColumnFamilyModel, self).save(
            *args,
            **kwargs
        )

        if hasattr(
            self._meta.model.objects,
            'denormalized_models'
        ):
            denormalized_models = self._meta.model.objects.denormalized_models

        else:
            denormalized_models = []

        if (
            None is denormalized_models or
            0 >= len(denormalized_models)
        ):
            return

        for model in denormalized_models:
            app_label = None

            if isinstance(model, tuple):
                app_label = model[0]
                model = model[1]

            if isinstance(model, str):
                if None is app_label:
                    app_label = self._meta.app_label

                model = apps.get_model(
                    app_label=app_label,
                    model_name=model
                )

            if (
                not issubclass(model, DjangoModel) or
                isinstance(self, model)
            ):
                continue

            if hasattr(model, 'should_denormalize'):
                if not model.should_denormalize(self):
                    continue

            denormalized_instance = model()
            if hasattr(model, 'denormalize'):
                denormalized_instance.denormalize(
                    self
                )

            else:
                ColumnFamilyManager.denormalize(
                    self,
                    denormalized_instance
                )

            super(ColumnFamilyModel, denormalized_instance).save(
                *args,
                **kwargs
            )

    def delete(
        self,
        *args,
        **kwargs
    ):
        if hasattr(
            self._meta.model.objects,
            'denormalized_models'
        ):
            denormalized_models = self._meta.model.objects.denormalized_models

        else:
            denormalized_models = []

        if (
            None is denormalized_models or
            0 >= len(denormalized_models)
        ):
            super(ColumnFamilyModel, self).delete(
                *args,
                **kwargs
            )
            return

        for model in denormalized_models:
            app_label = None

            if isinstance(model, tuple):
                app_label = model[0]
                model = model[1]

            if isinstance(model, str):
                if None is app_label:
                    app_label = self._meta.app_label

                model = apps.get_model(
                    app_label=app_label,
                    model_name=model
                )

            if (
                not issubclass(model, DjangoModel) or
                isinstance(self, model)
            ):
                continue

            if hasattr(model, 'should_denormalize'):
                if not model.should_denormalize(self):
                    continue

            denormalized_instance = model()
            if hasattr(model, 'denormalize'):
                denormalized_instance.denormalize(
                    self
                )

            else:
                ColumnFamilyManager.denormalize(
                    self,
                    denormalized_instance
                )

            super(
                ColumnFamilyModel,
                denormalized_instance
            ).delete(
                *args,
                **kwargs
            )

        super(ColumnFamilyModel, self).delete(
            *args,
            **kwargs
        )

    objects = ColumnFamilyManager()

    class PrimaryKeyInconsistencyException(Exception):
        pass
